def create_subscriber(topic, msg_type, callback, queue_size=10):
    """
    Create a ROS subscriber.

    :param topic: The name of the topic to subscribe to.
    :param msg_type: The type of message to subscribe to.
    :param callback: The callback function to call when a message is received.
    :param queue_size: The size of the message queue.
    :return: A ROS subscriber object.
    """
    # This is a placeholder implementation. In a real implementation, you would use
    # the appropriate ROS API to create the subscriber.
    raise NotImplementedError("ROS is not available in this environment. ")