def create_action_client(topic_name: str, action_message) -> None:
    """
    Create an action client for the given topic name and action message type.

    Args:
        topic_name (str): The name of the action topic.
        action_message: The action message type.

    Returns:
        None
    """
    raise NotImplementedError("No ROS environment detected.")