import unittest


class TestGiskardInterface(unittest.TestCase):
    def test_something(self):

        self.assertEqual(True, True)  # add assertion here


if __name__ == '__main__':
    unittest.main()
