from pycrap.ontologies.crax.classes import *


class BowlPreferredGraspAlignment(PreferredGraspAlignment):
    ...

class DefaultPreferredGraspAlignment(PreferredGraspAlignment):
    ...

class SpoonPreferredGraspAlignment(PreferredGraspAlignment):
    ...

class CerealPreferredGraspAlignment(PreferredGraspAlignment):
    ...


